/*======================================================================*\
|*		Unicode character conversion functions			*|
|*		substituting wide character functions			*|
|*			wc*, wcs*, mb*					*|
|*		and nl_ langinfo					*|
|*		for systems without proper locale support		*|
\*======================================================================*/

#include "uwc.h"
#include "wcwidth.c"
#include <string.h>
#include <locale.h>
#include <stdlib.h>


/*======================================================================*\
|*			Debugging					*|
\*======================================================================*/

#define dont_debug

#ifdef dont_debug
#undef debug
#endif

#ifdef debug
#include <stdio.h>
#define trace(o)	printf o
#else
#define trace(o)
#endif


/*======================================================================*\
|*			UTF-8 auxiliary functions			*|
\*======================================================================*/

static
int utf8_encode (unsigned long unichar, char * buf)
{
  int len;

  if (unichar < 0x80) {
	len = 1;
	* buf ++ = unichar;
  } else if (unichar < 0x800) {
	len = 2;
	* buf ++ = 0xC0 | (unichar >> 6);
	* buf ++ = 0x80 | (unichar & 0x3F);
  } else if (unichar < 0x10000) {
	len = 3;
	* buf ++ = 0xE0 | (unichar >> 12);
	* buf ++ = 0x80 | ((unichar >> 6) & 0x3F);
	* buf ++ = 0x80 | (unichar & 0x3F);
  } else if (unichar < 0x200000) {
	len = 4;
	* buf ++ = 0xF0 | (unichar >> 18);
	* buf ++ = 0x80 | ((unichar >> 12) & 0x3F);
	* buf ++ = 0x80 | ((unichar >> 6) & 0x3F);
	* buf ++ = 0x80 | (unichar & 0x3F);
  } else if (unichar < 0x4000000) {
	len = 5;
	* buf ++ = 0xF8 | (unichar >> 24);
	* buf ++ = 0x80 | ((unichar >> 18) & 0x3F);
	* buf ++ = 0x80 | ((unichar >> 12) & 0x3F);
	* buf ++ = 0x80 | ((unichar >> 6) & 0x3F);
	* buf ++ = 0x80 | (unichar & 0x3F);
  } else if (unichar < 0x80000000) {
	len = 6;
	* buf ++ = 0xFC | (unichar >> 30);
	* buf ++ = 0x80 | ((unichar >> 24) & 0x3F);
	* buf ++ = 0x80 | ((unichar >> 18) & 0x3F);
	* buf ++ = 0x80 | ((unichar >> 12) & 0x3F);
	* buf ++ = 0x80 | ((unichar >> 6) & 0x3F);
	* buf ++ = 0x80 | (unichar & 0x3F);
  } else {
	len = -1;
  }
  * buf = '\0';
  return len;
}

static
int utf8_len (unsigned char c)
{
	if ((c & 0x80) == 0x00) {
		return 1;
	} else if ((c & 0xE0) == 0xC0) {
		return 2;
	} else if ((c & 0xF0) == 0xE0) {
		return 3;
	} else if ((c & 0xF8) == 0xF0) {
		return 4;
	} else if ((c & 0xFC) == 0xF8) {
		return 5;
	} else if ((c & 0xFE) == 0xFC) {
		return 6;
	} else { /* illegal UTF-8 code */
		return 1;
	}
}

static
void utf8_info (const char * u, int * length, wchar_t * ucs)
{
  char * textpoi = (char *) u;
  unsigned char c = * textpoi;
  int utfcount;
  unsigned long unichar;

	if (c == 0x00) {
		utfcount = 0;
		unichar = 0;
	} else if ((c & 0x80) == 0x00) {
		utfcount = 1;
		unichar = c;
	} else if ((c & 0xE0) == 0xC0) {
		utfcount = 2;
		unichar = c & 0x1F;
	} else if ((c & 0xF0) == 0xE0) {
		utfcount = 3;
		unichar = c & 0x0F;
	} else if ((c & 0xF8) == 0xF0) {
		utfcount = 4;
		unichar = c & 0x07;
	} else if ((c & 0xFC) == 0xF8) {
		utfcount = 5;
		unichar = c & 0x03;
	} else if ((c & 0xFE) == 0xFC) {
		utfcount = 6;
		unichar = c & 0x01;
	} else if (c == 0xFE) {
		/* illegal UTF-8 code */
		utfcount = 1;
		unichar = '4';

		* length = -1;
		return;
	} else if (c == 0xFF) {
		/* illegal UTF-8 code */
		utfcount = 1;
		unichar = '5';

		* length = -1;
		return;
	} else {
		/* illegal UTF-8 sequence character */
		utfcount = 1;
		unichar = '8';

		* length = -1;
		return;
	}

	* length = utfcount;

	utfcount --;
	textpoi ++;
	while (utfcount > 0 && (* textpoi & 0xC0) == 0x80) {
		unichar = (unichar << 6) | (* textpoi & 0x3F);
		utfcount --;
		textpoi ++;
	}
	if (utfcount > 0) {
		/* too short UTF-8 sequence */
		unichar = (unsigned char) '';
		* length -= utfcount;

		* length = -1;
		return;
	}

	* ucs = unichar;
	if (* ucs != unichar) {
		/* wchar_t not wide enough (e.g. cygwin) */
		trace (("utf8_info %04lX -> %04X\n", unichar, * ucs));
		* ucs = 0xFFFD;	/* REPLACEMENT CHARACTER */
	}
}


/*======================================================================*\
|*			UTF-8 mode detection				*|
\*======================================================================*/

static int utf8_mode_checked = 0;
static int utf8_mode = 1;

/**
   Get locale category value for given locale variable name from environment.
 */
static
char * locale (char * varname)
{
    char * locale = getenv ("LC_ALL");
    if (! locale || ! * locale) {
	locale = getenv (varname);
	if (! locale || ! * locale) {
		locale = getenv ("LANG");
		if (! locale) {
			locale = "";
		}
	}
    }
    return locale;
}

/**
   List of known UTF-8 locales without ".UTF-8" or ".utf8" suffix.
 */
static char * utf8_locales [] = {
	"aa_ER",
	"aa_ER@saaho",
	"aa_ET",
	"am_ET",
	"ar_IN",
	"bn_BD",
	"bn_IN",
	"byn_ER",
	"en_IN",
	"fa_IR",
	"gez_ER",
	"gez_ER@abegede",
	"gez_ET",
	"gez_ET@abegede",
	"gu_IN",
	"hi_IN",
	"kn_IN",
	"lo_LA",
	"ml_IN",
	"mn_MN",
	"mr_IN",
	"ne_NP",
	"om_ET",
	"pa_IN",
	"se_NO",
	"sid_ET",
	"so_ET",
	"ta_IN",
	"te_IN",
	"ti_ER",
	"ti_ET",
	"tig_ER",
	"ur_PK",
	"uz_UZ@cyrillic",
	"vi_VN",
	0
};

/**
   Check if locale without explicit suffix is known as UTF-8 encoded.
 */
static
int
in_utf8_locales (__const char * lc)
{
    char * * lpoi = utf8_locales;
    while (* lpoi) {
	if (strcmp (* lpoi, lc) == 0) {
		return 1;
	}
	lpoi ++;
    }
    return 0;
}

/**
   Determine if locale is UTF-8 encoded.
 */
static
void check_utf8_mode (__const char * lc_ctype)
{
    if (! * lc_ctype) {
	/* if given locale is empty, take locale from environment */
	lc_ctype = locale ("LC_CTYPE");
    }
    utf8_mode = strstr (lc_ctype, ".UTF-8") || strstr (lc_ctype, ".utf8")
		|| in_utf8_locales (lc_ctype);
    utf8_mode_checked = 1;
}

int is_u_utf8_mode ()
{
    if (! utf8_mode_checked) {
	check_utf8_mode ("");
    }
    return utf8_mode;
}

#define locale_name_len 99
static char locale_name [locale_name_len];

/* Set and/or return the current locale.  */
char * u_setlocale (int __category, __const char * __locale)
{
	char * loc = setlocale (__category, __locale);
	if (__category == LC_ALL || __category == LC_CTYPE) {
		char * codeset = nl_langinfo (CODESET);
		if (strcmp (codeset, "UTF-8") == 0) {
			trace (("setlocale %d <%s> (UTF-8) -> <%s>\n", __category, __locale, loc));
			return loc;
		} else {
			if (__locale) {
				check_utf8_mode (__locale);
			} else {
				/* todo: cache previous result */
				if (utf8_mode) {
					return "en_US.UTF-8";
				} else {
					return "C";
				}
			}
			if (utf8_mode) {
				/* append ".UTF-8" to return value */
				char * dot = strchr (__locale, '.');
				int len;
				if (dot) {
					len = dot - __locale;
				} else {
					len = strlen (__locale);
				}
				if (len + 6 >= locale_name_len) {
					len = locale_name_len - 7;
				}
				locale_name [0] = '\0';
				strncat (locale_name, __locale, len);
				strcat (locale_name, ".UTF-8");
				trace (("setlocale %d <%s> -> <%s> -> <%s>\n", __category, __locale, loc, locale_name));
				return locale_name;
			} else {
				trace (("setlocale %d <%s> (! UTF-8) -> <%s>\n", __category, __locale, loc));
				return loc;
			}
		}
	} else {
		trace (("setlocale %d <%s> -> <%s>\n", __category, __locale, loc));
		return loc;
	}
}

/* Return the current locale's value for ITEM.
   If ITEM is invalid, an empty string is returned.

   The string returned will not change until `setlocale' is called;
   it is usually in read-only memory and cannot be modified.  */
char * u_nl_langinfo (nl_item __item)
{
    if (__item == CODESET) {
	if (is_u_utf8_mode ()) {
		return "UTF-8";
	} else {
		return nl_langinfo (__item);
	}
    } else {
	return nl_langinfo (__item);
    }
}


/*======================================================================*\
|*			X input functions				*|
\*======================================================================*/

/* (No description in Xlib.h) */
int u_XwcLookupString (XIC ic, XKeyPressedEvent *event, 
                     wchar_t * buffer_return, int wchars_buffer, 
                     KeySym * keysym_return, Status * status_return)
{
#define ubuflen 999
    if (is_u_utf8_mode ()) {
	char ubuf [ubuflen];
#ifdef X_HAVE_UTF8_STRING
	int ulen = Xutf8LookupString (ic, event, 
				ubuf, ubuflen, 
				keysym_return, status_return);
#else
	???
#endif
	return u_mbstowcs (buffer_return, ubuf, wchars_buffer);
    } else {
	return XwcLookupString (ic, event, 
				buffer_return, wchars_buffer, 
				keysym_return, status_return);
    }
}


/*======================================================================*\
|*			Conversion functions				*|
\*======================================================================*/

/* Return the length of the multibyte character
   in S, which is no longer than N.  */
int u_mblen (__const char * __s, size_t __n)
{
    if (is_u_utf8_mode ()) {
	return u_mbtowc (0, __s, __n);
    } else {
	return mblen (__s, __n);
    }
}

/* Return number of bytes in multibyte character pointed to by S.  */
size_t u_mbrlen (__const char * __restrict __s, size_t __n,
		      mbstate_t * __restrict __ps)
{
    if (is_u_utf8_mode ()) {
	return u_mblen (__s, __n);
    } else {
	return mbrlen (__s, __n, __ps);
    }
}


#define dont_debug_state
#define store_incomplete

/* Return the length of the given multibyte character,
   putting its `wchar_t' representation in *PWC.  */
int u_mbtowc (wchar_t * __restrict __pwc,
		__const char * __restrict __s, size_t __n)
{
    if (is_u_utf8_mode ()) {
	if (__s) {
		int len;
		wchar_t wc;
		static char s0 [15] = "";
		__const char * spoi;
		int maxlen = __n;
		int len0 = 0;
#ifdef debug
		wc = 0x8888;
#endif

		/* if incomplete UTF-8 remains from previous call, combine */
		if (* s0) {
			int maxcop = 6;
			if (__n < maxcop) {
				maxcop = __n;
			}

			len0 = strlen (s0);

			strncat (s0, __s, maxcop);
			maxlen = strlen (s0);
#ifdef debug_state
			{ int i;
				printf ("u_mbtowc combines");
				for (i = 0; s0 [i]; i ++) {
					printf (" %02X", (unsigned char) s0 [i]);
				}
				printf ("\n");
			}
#endif
			spoi = s0;
		} else {
			spoi = __s;
		}

		if (utf8_len (* spoi) > maxlen) {
			/* remember UTF-8 prefix for next call */
			if (spoi != s0) {
				s0 [0] = '\0';
#ifdef store_incomplete
				strncat (s0, __s, __n);
#endif
			}
#ifdef debug
			{ int i;
				printf ("u_mbtowc");
				for (i = 0; i <= __n && i < 8; i ++) {
					printf (" %02X", (unsigned char) __s [i]);
				}
				printf (" (%d) -> ()\n", __n);
			}
#endif
#ifdef debug_state
			{ int i;
				printf ("u_mbtowc stores");
				for (i = 0; s0 [i]; i ++) {
					printf (" %02X", (unsigned char) s0 [i]);
				}
				printf ("\n");
			}
#endif
#ifdef store_incomplete
			return -2;
#else
			return -1;
#endif
		}

		utf8_info (spoi, & len, & wc);
		s0 [0] = '\0';
#ifdef debug
		{ int i;
			printf ("u_mbtowc");
			for (i = 0; i <= __n && i < 8; i ++) {
				printf (" %02X", (unsigned char) __s [i]);
			}
		}
#endif
		if (len < 0) {
			trace ((" (%d) -> (%d) %02X -> -1\n", __n, len, wc));
			return -1;
		} else if (len <= __n) {
			if (__pwc) {
				* __pwc = wc;
			}
			trace ((" (%d) -> (%d) %02X -> %d\n", __n, len, wc, len - len0));
			return len - len0;
		} else {
			trace ((" (%d) -> (%d) %02X -> -1\n", __n, len, wc));
			return -1;
		}
	} else {
		return u_mbsinit (0);
	}
    } else {
	return mbtowc (__pwc, __s, __n);
    }
}

/* Write wide character representation of multibyte character pointed
   to by S to PWC.  */
size_t u_mbrtowc (wchar_t * __restrict __pwc,
		       __const char * __restrict __s, size_t __n,
		       mbstate_t * __p)
{
    if (is_u_utf8_mode ()) {
	return u_mbtowc (__pwc, __s, __n);
    } else {
	return mbrtowc (__pwc, __s, __n, __p);
    }
}

/* Put the multibyte character represented
   by WCHAR in S, returning its length.  */
int u_wctomb (char * __s, wchar_t __wchar)
{
    if (is_u_utf8_mode ()) {
	if (__s) {
		int len = utf8_encode (__wchar, __s);
		trace (("u_wctomb %04X -> %d\n", __wchar, len));
		return len;
	} else {
		return u_mbsinit (0);
	}
    } else {
	return wctomb (__s, __wchar);
    }
}

/* Write multibyte representation of wide character WC to S.  */
size_t u_wcrtomb (char * __restrict __s, wchar_t __wc,
		       mbstate_t * __restrict __ps)
{
    if (is_u_utf8_mode ()) {
	return u_wctomb (__s, __wc);
    } else {
	return wcrtomb (__s, __wc, __ps);
    }
}


/* Convert a multibyte string to a wide char string.  */
size_t u_mbstowcs (wchar_t * __restrict  __pwcs,
			__const char * __restrict __s, size_t __n)
{
    if (is_u_utf8_mode ()) {
	wchar_t wc;
	wchar_t * wpoi = __pwcs;
	int count = 0;
	if (! __pwcs) {
		wpoi = & wc;
	}
	while (count < __n) {
		int len;
		utf8_info (__s, & len, wpoi);
		if (len < 0) {
			trace (("u_mbstowcs %04X -> -1\n", * wpoi, len));
			return -1;
		} else if (* wpoi == 0) {
			trace (("u_mbstowcs 0 -> %d\n", count));
			return count;
		}
		count ++;
		__s += len;
		if (__pwcs) {
			wpoi ++;
		}
	}
	trace (("u_mbstowcs %04X -> %d\n", * wpoi, __n));
	return __n;
    } else {
	return (__pwcs, __s, __n);
    }
}

/* Convert a wide char string to multibyte string.  */
size_t u_wcstombs (char * __restrict __s,
			__const wchar_t * __restrict __pwcs, size_t __n)
{
    if (is_u_utf8_mode ()) {
	int length = 0;
	while (* __pwcs) {
		char buf [7];
		int len = utf8_encode (* __pwcs, buf);
		if (len < 0) {
			length = -1;
			break;
		} else if (length + len + 1 > __n) {
			break;
		} else {
			if (__s) {
				(void) strcpy (__s, buf);
				__s += len;
			}
			length += len;
			__pwcs ++;
		}
	}
	if (__s) {
		* __s = '\0';
	}
	trace (("u_wcstombs ... %04X -> %d\n", * __pwcs, length));
	return length;
    } else {
	return wcstombs (__s, __pwcs, __n);
    }
}


/* Determine whether PS points to an object representing the initial
   state.  */
int u_mbsinit (__const mbstate_t * __ps)
{
    if (is_u_utf8_mode ()) {
	return 0;
    } else {
	return mbsinit (__ps);
    }
}

/* Return the number of wide characters in S.  */
size_t u_wcslen (__const wchar_t * __s)
{
    if (is_u_utf8_mode ()) {
	int i = 0;
	while (* __s ++ != 0) {
		i ++;
	}
	return i;
    } else {
	return wcslen (__s);
    }
}


/* Determine number of column positions required for C.  */
int u_wcwidth (wchar_t __c)
{
    if (is_u_utf8_mode ()) {
	return mk_wcwidth (__c);
    } else {
	return wcwidth (__c);
    }
}

/* Determine number of column positions required for first N wide
   characters (or fewer if S ends before this) in S.  */
int u_wcswidth (__const wchar_t * __s, size_t __n)
{
    if (is_u_utf8_mode ()) {
	return mk_wcswidth (__s, __n);
    } else {
	return wcswidth (__s, __n);
    }
}


/*======================================================================*\
|*			End						*|
\*======================================================================*/
